from .onnx_to_trt import Onnx2Trt
from .saliency_segment_trt import SaliencySegmentTrt
from .object_classification_trt import ObjectClassificationTrt
from .object_detection_trt import ObjectDetectionTrt
from .instance_segment_trt import InstanceSegmentTrt



import cv2
import numpy as np
import tensorrt as trt
# import pycuda.autoinit
import pycuda.driver as cuda
from .general import *


class BaseEngine(object):
    def __init__(self, trt_file, image_size=(640, 640)):
        self.image_size = image_size
        self.mean = None
        self.std = None
        self.n_classes = 80
        self.class_names = ['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat',
                            'traffic light',
                            'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse',
                            'sheep', 'cow',
                            'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie',
                            'suitcase', 'frisbee',
                            'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat', 'baseball glove', 'skateboard',
                            'surfboard',
                            'tennis racket', 'bottle', 'wine glass', 'cup', 'fork', 'knife', 'spoon', 'bowl', 'banana',
                            'apple',
                            'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza', 'donut', 'cake', 'chair',
                            'couch',
                            'potted plant', 'bed', 'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
                            'keyboard', 'cell phone',
                            'microwave', 'oven', 'toaster', 'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors',
                            'teddy bear',
                            'hair drier', 'toothbrush']

        cuda.init()
        self.cfx = cuda.Device(0).make_context()

        logger = trt.Logger(trt.Logger.WARNING)
        runtime = trt.Runtime(logger)
        trt.init_libnvinfer_plugins(logger, '')
        with open(trt_file, "rb") as f:
            serialized_engine = f.read()
        engine = runtime.deserialize_cuda_engine(serialized_engine)
        self.context = engine.create_execution_context()
        self.inputs, self.outputs, self.bindings = [], [], []
        self.stream = cuda.Stream()
        for binding in engine:
            size = trt.volume(engine.get_binding_shape(binding))
            dtype = trt.nptype(engine.get_binding_dtype(binding))
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            self.bindings.append(int(device_mem))
            if engine.binding_is_input(binding):
                self.inputs.append({'host': host_mem, 'device': device_mem})
            else:
                self.outputs.append({'host': host_mem, 'device': device_mem})

    def __del__(self):
        self.cfx.detach()

    def infer(self, img):
        self.cfx.push()

        self.inputs[0]['host'] = np.ravel(img)
        # transfer data to the gpu
        for inp in self.inputs:
            cuda.memcpy_htod_async(inp['device'], inp['host'], self.stream)
        # run inference
        self.context.execute_async_v2(
            bindings=self.bindings,
            stream_handle=self.stream.handle)
        # fetch outputs from gpu
        for out in self.outputs:
            cuda.memcpy_dtoh_async(out['host'], out['device'], self.stream)
        # synchronize stream
        self.stream.synchronize()
        data = [out['host'] for out in self.outputs]

        self.cfx.pop()
        return data

    def detect_video(self, video_path, conf=0.5, iou=0.45, end2end=False):
        cap = cv2.VideoCapture(video_path)
        fourcc = cv2.VideoWriter_fourcc(*'XVID')
        fps = int(round(cap.get(cv2.CAP_PROP_FPS)))
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        out = cv2.VideoWriter('results.avi', fourcc, fps, (width, height))
        fps = 0
        import time
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            blob, ratio = preproc(frame, self.image_size, self.mean, self.std)
            t1 = time.time()
            data = self.infer(blob)
            fps = (fps + (1. / (time.time() - t1))) / 2
            frame = cv2.putText(frame, "FPS:%d " % fps, (0, 40), cv2.FONT_HERSHEY_SIMPLEX, 1,
                                (0, 0, 255), 2)
            if end2end:
                num, final_boxes, final_scores, final_cls_inds = data
                final_boxes = np.reshape(final_boxes / ratio, (-1, 4))
                dets = np.concatenate([final_boxes[:num[0]], np.array(final_scores)[:num[0]].reshape(-1, 1),
                                       np.array(final_cls_inds)[:num[0]].reshape(-1, 1)], axis=-1)
            else:
                predictions = np.reshape(data, (1, -1, int(5 + self.n_classes)))[0]
                dets = self.postprocess(predictions, ratio, conf, iou)

            if dets is not None:
                final_boxes, final_scores, final_cls_inds = dets[:, :4], dets[:, 4], dets[:, 5]
                frame = vis(frame, final_boxes, final_scores, final_cls_inds,
                            conf=conf, class_names=self.class_names)
            cv2.imshow('frame', frame)
            out.write(frame)
            if cv2.waitKey(25) & 0xFF == ord('q'):
                break
        out.release()
        cap.release()
        cv2.destroyAllWindows()

    def detect(self, origin_img, conf=0.5, iou=0.45, end2end=False):
        img, ratio = preproc(origin_img, self.image_size, self.mean, self.std)
        data = self.infer(img)
        if end2end:
            num, final_boxes, final_scores, final_cls_inds = data
            final_boxes = np.reshape(final_boxes / ratio, (-1, 4))
            dets = np.concatenate([final_boxes[:num[0]], np.array(final_scores)[:num[0]].reshape(-1, 1),
                                   np.array(final_cls_inds)[:num[0]].reshape(-1, 1)], axis=-1)
        else:
            predictions = np.reshape(data, (1, -1, int(5 + self.n_classes)))[0]
            dets = self.postprocess(predictions, ratio, conf, iou)

        return dets

    def segment(self, origin_img, conf=0.5, iou=0.45, end2end=False):
        img, ratio = preproc(origin_img, self.image_size, self.mean, self.std)
        data = self.infer(img)[-1]

        predictions = np.reshape(data, (-1, 38))[:, :6]
        dets = self.postprocess(predictions, ratio, conf, iou)

        return dets

    @staticmethod
    def postprocess(predictions, ratio, score_thr, nms_thr):
        boxes = predictions[:, :4]
        scores = predictions[:, 4:5] * predictions[:, 5:]
        boxes_xyxy = np.ones_like(boxes)
        boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.
        boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.
        boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.
        boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.
        boxes_xyxy /= ratio
        dets = multiclass_nms(boxes_xyxy, scores, nms_thr=nms_thr, score_thr=score_thr)
        return dets



import cv2
import numpy as np
from ..utils import sigmoid
from ..utils import BaseEngine


def preprocess(image, input_size, mean, std, swap=(2, 0, 1)):
    resized_img = cv2.resize(image, (input_size[1], input_size[0]), interpolation=cv2.INTER_LINEAR)
    resized_img = resized_img[:, :, ::-1].astype(np.float32)
    resized_img /= 255.0
    if mean is not None:
        resized_img -= np.array(mean)
    if std is not None:
        resized_img /= np.array(std)
    resized_img = resized_img.transpose(swap)
    resized_img = np.ascontiguousarray(resized_img, dtype=np.float32)
    return resized_img


class SegmentTrtPredictor(BaseEngine):
    def __init__(self, trt_file, image_size=(640, 640)):
        super(SegmentTrtPredictor, self).__init__(trt_file)
        self.image_size = image_size  # (h, w)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]

    def predict(self, origin_img, class_names: list):
        n_classes = len(class_names)
        h, w = origin_img.shape[:2]
        img = preprocess(origin_img, self.image_size, self.mean, self.std)
        data = self.infer(img)[0]   # fix 1 -> 0

        pred = np.reshape(data, (-1, n_classes, self.image_size[0], self.image_size[1]))
        pred = sigmoid(pred[0, 0])
        pred = np.asarray(pred * 255).astype(np.uint8)

        pred = cv2.resize(pred, (w, h), interpolation=cv2.INTER_LINEAR)
        return pred



import numpy as np
from ..utils import preproc, multiclass_nms
from ..utils import BaseEngine


class DetectionTrtPredictor(BaseEngine):
    def __init__(self, trt_file, image_size=(640, 640)):
        super(DetectionTrtPredictor, self).__init__(trt_file)
        self.image_size = image_size  # (h, w)
        # self.mean = [0.485, 0.456, 0.406]
        # self.std = [0.229, 0.224, 0.225]

    def predict(self, origin_img, class_names, use_preprocess, conf, iou, end2end=False):
        n_classes = len(class_names)
        img, ratio = preproc(origin_img, self.image_size, self.mean, self.std) if use_preprocess else (origin_img, 1.0)
        data = self.infer(img)
        if end2end:
            num, final_boxes, final_scores, final_cls_inds = data
            final_boxes = np.reshape(final_boxes / ratio, (-1, 4))
            dets = np.concatenate([final_boxes[:num[0]], np.array(final_scores)[:num[0]].reshape(-1, 1),
                                   np.array(final_cls_inds)[:num[0]].reshape(-1, 1)], axis=-1)
        else:
            predictions = np.reshape(data, (1, -1, int(5 + n_classes)))[0]
            dets = self.postprocess(predictions, ratio, conf, iou)

        return dets

    @staticmethod
    def postprocess(predictions, ratio, score_thr, nms_thr):
        boxes = predictions[:, :4]
        scores = predictions[:, 4:5] * predictions[:, 5:6]
        boxes_xyxy = np.ones_like(boxes)
        boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.
        boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.
        boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.
        boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.
        boxes_xyxy /= ratio
        dets = multiclass_nms(boxes_xyxy, scores, nms_thr=nms_thr, score_thr=score_thr)
        return dets



import numpy as np
from ..model_zoo import SegmentTrtPredictor


class SaliencySegmentTrt:
    def __init__(self,
                 trt_file="models/model-32.trt",
                 input_shape=(352, 352),
                 conf=0.01,
                 iou=0.45,
                 end2end=False):
        self.conf = conf
        self.iou = iou
        self.end2end = end2end
        self.model = SegmentTrtPredictor(trt_file=trt_file, image_size=input_shape)

    def predict(self, image, class_names):
        return self.model.predict(image, class_names)

    @staticmethod
    def show(image, mask):
        result = image * (mask[:, :, np.newaxis] / 255)
        return result.astype(np.uint8)



import time
import cv2
from ..model_zoo import InstanceTrtPredictor


class InstanceSegmentTrt:
    def __init__(self,
                 trt_file="models/yolov7-seg.trt",
                 input_shape=(640, 640),
                 conf=0.01,
                 iou=0.45,
                 use_mask=False,
                 use_cupy=False,
                 multi_label=False):
        self.conf = conf
        self.iou = iou
        self.use_mask = use_mask
        self.use_cupy = use_cupy
        self.multi_label = multi_label
        self.model = InstanceTrtPredictor(trt_file=trt_file, image_size=input_shape)

    def predict(self, image, class_names, use_preprocess=False):
        return self.model.predict(image, class_names, use_preprocess, self.conf, self.iou, self.use_mask,
                                  self.use_cupy, self.multi_label)

    @staticmethod
    def show(image, results):
        if results is None or len(results) == 0:
            return image

        for i, result in enumerate(results):
            x1, y1, x2, y2, x3, y3, x4, y4, score, label = result
            cv2.line(image, pt1=(x1, y1), pt2=(x2, y2), color=(255, 255, 0), thickness=10)
            cv2.line(image, pt1=(x2, y2), pt2=(x3, y3), color=(255, 255, 0), thickness=10)
            cv2.line(image, pt1=(x3, y3), pt2=(x4, y4), color=(255, 255, 0), thickness=10)
            cv2.line(image, pt1=(x4, y4), pt2=(x1, y1), color=(255, 255, 0), thickness=10)
            cv2.putText(image, 'id: %d, score: %.2f' % (label, score),
                        (x1, y1 - 4), cv2.FONT_HERSHEY_PLAIN, 2, (0, 0, 255), thickness=2)
        return image

    def predict_video(self, file_path=0, class_names=['person'], show=False):
        capture = cv2.VideoCapture(file_path)
        if capture.isOpened():
            if show:
                capture.set(cv2.CAP_PROP_FRAME_WIDTH, 640)
                capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 480)
            frame_id = 0
            while True:
                read_code, frame = capture.read()
                if not read_code: break
                s = time.time()
                results = self.predict(frame, class_names)
                print("object nums: ", frame_id, len(results) if results is not None else 0, time.time() - s)
                frame = self.show(frame, results)
                if show:
                    cv2.imshow("object_detect", frame)
                    if cv2.waitKey(1) == ord('q'):
                        break
                else:
                    cv2.imwrite('images/' + str(frame_id) + '.jpg', frame)
                frame_id += 1
            capture.release()
            if show: cv2.destroyWindow("object_detect")



import cv2
import numpy as np
from ..utils import softmax
from ..utils import BaseEngine


def preprocess(image, input_size, mean, std, swap=(2, 0, 1)):
    # 传入图像默认为RGB
    padded_img = cv2.resize(image, input_size, interpolation=cv2.INTER_LINEAR)
    padded_img = padded_img.astype(np.float32)
    padded_img /= 255.0
    if mean is not None:
        padded_img -= np.array(mean)
    if std is not None:
        padded_img /= np.array(std)
    padded_img = padded_img.transpose(swap)
    padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
    return padded_img


class ClassificationTrtPredictor(BaseEngine):
    def __init__(self, trt_file, image_size=(224, 224)):
        super(ClassificationTrtPredictor, self).__init__(trt_file)
        self.image_size = image_size  # (h, w)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]

    def predict(self, origin_img, class_names: list, use_preprocess=False):
        if isinstance(origin_img, list):
            batch_size = len(origin_img)
            if use_preprocess:
                img = [preprocess(im, self.image_size, self.mean, self.std) for im in
                       origin_img] if use_preprocess else origin_img
            else:
                img = origin_img
        else:
            batch_size = 1
            img = preprocess(origin_img, self.image_size, self.mean, self.std) if use_preprocess else origin_img

        data = self.infer(np.array(img))[-1]
        predictions = np.reshape(data, (batch_size, -1))
        output_score = np.array([softmax(p) for p in predictions])
        output_class = np.argmax(output_score, axis=1)

        return [{'category_id': int(c),
                 'category': class_names[int(c)] if len(class_names) else str(c),
                 'score': float(s[int(c)])} for c, s in zip(output_class, output_score)]

    def feature(self, origin_img, num_classes=1000, use_preprocess=False):
        if isinstance(origin_img, list):
            batch_size = len(origin_img)
            if use_preprocess:
                img = [preprocess(im, self.image_size, self.mean, self.std) for im in
                       origin_img] if use_preprocess else origin_img
            else:
                img = origin_img
        else:
            batch_size = 1
            img = preprocess(origin_img, self.image_size, self.mean, self.std) if use_preprocess else origin_img
        data = self.infer(np.array(img))[0]
        predictions = np.reshape(data, (batch_size, -1))

        return predictions



from .trt_export import *
from .classification_trt_predictor import ClassificationTrtPredictor
from .detection_trt_predictor import DetectionTrtPredictor
from .instance_trt_predictor import InstanceTrtPredictor
from .segment_trt_predictor import SegmentTrtPredictor



import time
import cv2
import numpy as np
from ..model_zoo import DetectionTrtPredictor


class ObjectDetectionTrt:
    def __init__(self,
                 trt_file="models/yolov7.trt",
                 input_shape=(640, 640),
                 conf=0.01,
                 iou=0.45,
                 end2end=False):
        self.conf = conf
        self.iou = iou
        self.end2end = end2end
        self.model = DetectionTrtPredictor(trt_file=trt_file, image_size=input_shape)

    def predict(self, image, class_names, use_preprocess=False):
        return self.model.predict(image, class_names, use_preprocess, self.conf, self.iou, self.end2end)

    @staticmethod
    def show(image, results):
        if results is None or len(results) == 0:
            return image
        index = 1
        for (box, score) in zip(list(results[:, :-2].astype(int)), list(results[:, -2])):
            cv2.rectangle(image, (box[0], box[1]), (box[2], box[3]), (255, 0, 255), 2)
            cv2.putText(image, 'id: %d, score: %.2f' % (index, score),
                        (box[0], box[1] - 4), cv2.FONT_HERSHEY_PLAIN, 2, (0, 0, 255), thickness=2)
            index += 1
        return image

    def predict_video(self, file_path=0, class_names=['person'], show=False):
        capture = cv2.VideoCapture(file_path)
        if capture.isOpened():
            if show:
                capture.set(cv2.CAP_PROP_FRAME_WIDTH, 640)
                capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 480)
            frame_id = 0
            while True:
                read_code, frame = capture.read()
                if not read_code: break
                s = time.time()
                results = self.predict(frame, class_names)
                print("object nums: ", frame_id, len(results) if results is not None else 0, time.time() - s)
                frame = self.show(frame, results)
                if show:
                    cv2.imshow("object_detect", frame)
                    if cv2.waitKey(1) == ord('q'):
                        break
                else:
                    cv2.imwrite('images/' + str(frame_id) + '.jpg', frame)
                frame_id += 1
            capture.release()
            if show: cv2.destroyWindow("object_detect")



import cv2
import numpy as np
from ..utils import preproc, multiclass_nms, singleclass_nms, sigmoid, fast_dot
from ..utils import BaseEngine
import pycuda.driver as cuda


class InstanceTrtPredictor(BaseEngine):
    def __init__(self, trt_file, image_size=(640, 640)):
        super(InstanceTrtPredictor, self).__init__(trt_file)
        self.image_size = image_size  # (h, w)
        self.mean = None
        self.std = None

    @staticmethod
    def get_rect(mask):
        h, w = mask.shape[:2]
        _, pred = cv2.threshold(mask, 128, 255, cv2.THRESH_BINARY)
        contours, _ = cv2.findContours(pred, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        if len(contours) < 1:
            return [[0, 0], [0, w - 1], [h - 1, w - 1], [h - 1, 0]]

        area = [cv2.contourArea(c) for c in contours]
        max_idx = np.argmax(area)
        min_rect = cv2.minAreaRect(contours[max_idx])
        min_rect = cv2.boxPoints(min_rect)

        return min_rect

    def predict(self, origin_img, class_names, use_preprocess, conf, iou, use_mask=False, use_cupy=False,
                multi_label=False):
        dets = []
        img, ratio = preproc(origin_img, self.image_size, self.mean, self.std) if use_preprocess else (origin_img, 1.0)
        data = self.infer(img)
        data = [np.ascontiguousarray(x, dtype=np.float64) for x in data]
        # 处理box
        mi = 5 + len(class_names)
        pred = np.reshape(data[-1], (1, -1, mi + 32))[0]
        boxes = self.postprocess(pred, ratio=1., score_thr=conf, nms_thr=iou, multi_label=multi_label)
        if boxes is None: return dets
        # 处理mask
        if use_mask:
            if use_cupy:
                import cupy as cp

                ctx = cuda.Device(0).retain_primary_context()
                ctx.push()
                gpu_data0 = cp.asarray(data[0])
                gpu_proto = cp.reshape(gpu_data0, (1, 32, -1))[0]
                gpu_boxes = cp.asarray(boxes[:, 6:])
                gpu_dot = cp.dot(gpu_boxes, gpu_proto)
                gpu_sigmoid_dot = 1 / (1 + cp.exp(-gpu_dot))
                gpu_masks = cp.reshape(gpu_sigmoid_dot, (-1, 160, 160))
                masks = cp.asnumpy(gpu_masks)
                ih, iw = 160 / self.image_size[0], 160 / self.image_size[1]
                ctx.pop()
            else:
                # csl: about 11ms on i7-8700
                proto = np.reshape(data[0], (1, 32, -1))[0]
                masks = np.reshape(sigmoid(fast_dot(boxes[:, 6:], proto)), (-1, 160, 160))
                ih, iw = 160 / self.image_size[0], 160 / self.image_size[1]

            for mask, box, score, label in zip(masks, boxes[:, :4], boxes[:, 4], boxes[:, 5]):
                x1, y1, x2, y2 = box[0] * iw, box[1] * ih, box[2] * iw, box[3] * ih
                crop = mask[int(y1):int(y2) + 1, int(x1):int(x2) + 1] * 255
                min_rect = self.get_rect(crop.astype(np.uint8))
                rect = [[(r[0] + x1) / ih / ratio, (r[1] + y1) / iw / ratio] for r in min_rect]
                dets.append([max(int(rect[0][0]), 0), max(int(rect[0][1]), 0),
                             max(int(rect[1][0]), 0), max(int(rect[1][1]), 0),
                             max(int(rect[2][0]), 0), max(int(rect[2][1]), 0),
                             max(int(rect[3][0]), 0), max(int(rect[3][1]), 0),
                             float(score), int(label)])
        else:
            for box, score, label in zip(boxes[:, :4], boxes[:, 4], boxes[:, 5]):
                x1, y1, x2, y2 = box[0] / ratio, box[1] / ratio, box[2] / ratio, box[3] / ratio
                dets.append([max(int(x1), 0), max(int(y1), 0),
                             max(int(x2), 0), max(int(y1), 0),
                             max(int(x2), 0), max(int(y2), 0),
                             max(int(x1), 0), max(int(y2), 0),
                             float(score), int(label)])

        return dets

    @staticmethod
    def postprocess(predictions, ratio, score_thr, nms_thr, multi_label=False):
        boxes = predictions[:, :4]
        scores = predictions[:, 4:5] * predictions[:, 5:-32]
        boxes_xyxy = np.ones_like(boxes)
        boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.
        boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.
        boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.
        boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.
        boxes_xyxy /= ratio

        if multi_label:
            dets = multiclass_nms(boxes_xyxy, scores, nms_thr=nms_thr, score_thr=score_thr, predictions=predictions)
        else:
            dets = singleclass_nms(boxes_xyxy, scores, nms_thr=nms_thr, score_thr=score_thr, predictions=predictions)

        return dets



import cv2
import numpy as np

_COLORS = np.array(
    [
        0.000, 0.447, 0.741,
        0.850, 0.325, 0.098,
        0.929, 0.694, 0.125,
        0.494, 0.184, 0.556,
        0.466, 0.674, 0.188,
        0.301, 0.745, 0.933,
        0.635, 0.078, 0.184,
        0.300, 0.300, 0.300,
        0.600, 0.600, 0.600,
        1.000, 0.000, 0.000,
        1.000, 0.500, 0.000,
        0.749, 0.749, 0.000,
        0.000, 1.000, 0.000,
        0.000, 0.000, 1.000,
        0.667, 0.000, 1.000,
        0.333, 0.333, 0.000,
        0.333, 0.667, 0.000,
        0.333, 1.000, 0.000,
        0.667, 0.333, 0.000,
        0.667, 0.667, 0.000,
        0.667, 1.000, 0.000,
        1.000, 0.333, 0.000,
        1.000, 0.667, 0.000,
        1.000, 1.000, 0.000,
        0.000, 0.333, 0.500,
        0.000, 0.667, 0.500,
        0.000, 1.000, 0.500,
        0.333, 0.000, 0.500,
        0.333, 0.333, 0.500,
        0.333, 0.667, 0.500,
        0.333, 1.000, 0.500,
        0.667, 0.000, 0.500,
        0.667, 0.333, 0.500,
        0.667, 0.667, 0.500,
        0.667, 1.000, 0.500,
        1.000, 0.000, 0.500,
        1.000, 0.333, 0.500,
        1.000, 0.667, 0.500,
        1.000, 1.000, 0.500,
        0.000, 0.333, 1.000,
        0.000, 0.667, 1.000,
        0.000, 1.000, 1.000,
        0.333, 0.000, 1.000,
        0.333, 0.333, 1.000,
        0.333, 0.667, 1.000,
        0.333, 1.000, 1.000,
        0.667, 0.000, 1.000,
        0.667, 0.333, 1.000,
        0.667, 0.667, 1.000,
        0.667, 1.000, 1.000,
        1.000, 0.000, 1.000,
        1.000, 0.333, 1.000,
        1.000, 0.667, 1.000,
        0.333, 0.000, 0.000,
        0.500, 0.000, 0.000,
        0.667, 0.000, 0.000,
        0.833, 0.000, 0.000,
        1.000, 0.000, 0.000,
        0.000, 0.167, 0.000,
        0.000, 0.333, 0.000,
        0.000, 0.500, 0.000,
        0.000, 0.667, 0.000,
        0.000, 0.833, 0.000,
        0.000, 1.000, 0.000,
        0.000, 0.000, 0.167,
        0.000, 0.000, 0.333,
        0.000, 0.000, 0.500,
        0.000, 0.000, 0.667,
        0.000, 0.000, 0.833,
        0.000, 0.000, 1.000,
        0.000, 0.000, 0.000,
        0.143, 0.143, 0.143,
        0.286, 0.286, 0.286,
        0.429, 0.429, 0.429,
        0.571, 0.571, 0.571,
        0.714, 0.714, 0.714,
        0.857, 0.857, 0.857,
        0.000, 0.447, 0.741,
        0.314, 0.717, 0.741,
        0.50, 0.5, 0
    ]
).astype(np.float32).reshape(-1, 3)


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def softmax(x):
    return np.exp(x) / np.sum(np.exp(x), axis=0)


def _force_forder(x):
    """
    Converts arrays x to fortran order. Returns
    a tuple in the form (x, is_transposed).
    """
    if x.flags.c_contiguous:
        return x.T, True
    else:
        return x, False


def fast_dot(A, B):
    """
    Uses blas libraries directly to perform dot product
    """
    from scipy import linalg

    a, trans_a = _force_forder(A)
    b, trans_b = _force_forder(B)
    gemm_dot = linalg.get_blas_funcs("gemm", arrays=(a, b))

    # gemm is implemented to compute: C = alpha * AB  + beta * C
    return gemm_dot(alpha=1.0, a=a, b=b, trans_a=trans_a, trans_b=trans_b)


def nms(boxes, scores, nms_thr):
    """Single class NMS implemented in Numpy."""
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= nms_thr)[0]
        order = order[inds + 1]

    return keep


def singleclass_nms(boxes, scores, nms_thr, score_thr, predictions=None):
    """Multiclass NMS implemented in Numpy"""
    num_classes = scores.shape[1]
    mi = 5 + num_classes  # mask start index

    cls_scores = np.max(scores, axis=1)
    cls_indexes = np.argmax(scores, axis=1)
    valid_score_mask = cls_scores > score_thr
    if valid_score_mask.sum() == 0:
        return None
    else:
        valid_indexes = cls_indexes[valid_score_mask]
        valid_scores = cls_scores[valid_score_mask]
        valid_boxes = boxes[valid_score_mask]
        valid_predictions = predictions[valid_score_mask]
        keep = nms(valid_boxes, valid_scores, nms_thr)
        if len(keep) > 0:
            if predictions is None:
                dets = np.concatenate([valid_boxes[keep], valid_scores[keep, None], valid_indexes[keep]], 1)
            else:
                valid_predictions[:, mi:] *= valid_predictions[:, 4:5]
                dets = np.concatenate(
                    [valid_boxes[keep], valid_scores[keep, None], valid_indexes[keep, None],
                     valid_predictions[keep, mi:]], 1)

    if len(dets) == 0:
        return None
    return dets


def multiclass_nms(boxes, scores, nms_thr, score_thr, predictions=None):
    """Multiclass NMS implemented in Numpy"""
    final_dets = []
    num_classes = scores.shape[1]
    mi = 5 + num_classes  # mask start index
    for cls_ind in range(num_classes):
        cls_scores = scores[:, cls_ind]
        valid_score_mask = cls_scores > score_thr
        if valid_score_mask.sum() == 0:
            continue
        else:
            valid_scores = cls_scores[valid_score_mask]
            valid_boxes = boxes[valid_score_mask]
            valid_predictions = predictions[valid_score_mask]
            keep = nms(valid_boxes, valid_scores, nms_thr)
            if len(keep) > 0:
                cls_inds = np.ones((len(keep), 1)) * cls_ind
                if predictions is None:
                    dets = np.concatenate([valid_boxes[keep], valid_scores[keep, None], cls_inds], 1)
                else:
                    valid_predictions[:, mi:] *= valid_predictions[:, 4:5]
                    dets = np.concatenate(
                        [valid_boxes[keep], valid_scores[keep, None], cls_inds, valid_predictions[keep, mi:]], 1)
                final_dets.append(dets)
    if len(final_dets) == 0:
        return None
    return np.concatenate(final_dets, 0)


def preproc(image, input_size, mean, std, swap=(2, 0, 1)):
    img = np.array(image)
    r = min(input_size[0] / img.shape[0], input_size[1] / img.shape[1])

    if img.shape[:2] == input_size:
        padded_img = img.astype(np.float32)
    else:
        # 传入图像默认为RGB
        if len(image.shape) == 3:
            padded_img = np.ones((input_size[0], input_size[1], 3)) * 114.0
        else:
            padded_img = np.ones(input_size) * 114.0
        resized_img = cv2.resize(
            img,
            (int(round(img.shape[1] * r)), int(round(img.shape[0] * r))),
            interpolation=cv2.INTER_LINEAR,
        ).astype(np.float32)
        padded_img[: int(round(img.shape[0] * r)), : int(round(img.shape[1] * r))] = resized_img
    padded_img /= 255.0
    if mean is not None:
        padded_img -= mean
    if std is not None:
        padded_img /= std
    padded_img = padded_img.transpose(swap)
    padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
    return padded_img, r


def preproc2(image, input_size, mean, std, swap=(2, 0, 1)):
    if len(image.shape) == 3:
        padded_img = np.ones((input_size[0], input_size[1], 3)) * 114.0
    else:
        padded_img = np.ones(input_size) * 114.0
    img = np.array(image)
    r = min(input_size[0] / img.shape[0], input_size[1] / img.shape[1])
    new_unpad = int(round(img.shape[1] * r)), int(round(img.shape[0] * r))
    dw, dh = input_size[1] - new_unpad[0], input_size[0] - new_unpad[1]
    dw /= 2
    dh /= 2
    resized_img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR).astype(np.float32)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    padded_img[top: input_size[1] - bottom, left: input_size[0] - right] = resized_img

    padded_img = padded_img[:, :, ::-1]
    padded_img /= 255.0
    if mean is not None:
        padded_img -= mean
    if std is not None:
        padded_img /= std
    padded_img = padded_img.transpose(swap)
    padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
    return padded_img, r


def preprocess(image, input_size, mean, std, swap=(2, 0, 1)):
    imh, imw = image.shape[:2]
    m = min(imh, imw)  # min dimension
    top, left = (imh - m) // 2, (imw - m) // 2
    padded_img = cv2.resize(image[top:top + m, left:left + m], input_size, interpolation=cv2.INTER_LINEAR)
    padded_img = padded_img[:, :, ::-1].astype(np.float32)
    padded_img /= 255.0
    if mean is not None:
        padded_img -= np.array(mean)
    if std is not None:
        padded_img /= np.array(std)
    padded_img = padded_img.transpose(swap)
    padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
    return padded_img


def vis(img, boxes, scores, cls_ids, conf=0.5, class_names=None):
    for i in range(len(boxes)):
        box = boxes[i]
        cls_id = int(cls_ids[i])
        score = scores[i]
        if score < conf:
            continue
        x0 = int(box[0])
        y0 = int(box[1])
        x1 = int(box[2])
        y1 = int(box[3])

        color = (_COLORS[cls_id] * 255).astype(np.uint8).tolist()
        text = '{}:{:.1f}%'.format(class_names[cls_id], score * 100)
        txt_color = (0, 0, 0) if np.mean(_COLORS[cls_id]) > 0.5 else (255, 255, 255)
        font = cv2.FONT_HERSHEY_SIMPLEX

        txt_size = cv2.getTextSize(text, font, 0.4, 1)[0]
        cv2.rectangle(img, (x0, y0), (x1, y1), color, 2)

        txt_bk_color = (_COLORS[cls_id] * 255 * 0.7).astype(np.uint8).tolist()
        cv2.rectangle(
            img,
            (x0, y0 + 1),
            (x0 + txt_size[0] + 1, y0 + int(1.5 * txt_size[1])),
            txt_bk_color,
            -1
        )
        cv2.putText(img, text, (x0, y0 + txt_size[1]), font, 0.4, txt_color, thickness=1)

    return img



from ..model_zoo import ClassificationTrtPredictor


class ObjectClassificationTrt:
    def __init__(self,
                 trt_file="models/person-cls.trt",
                 input_shape=(224, 224),
                 conf=0.01,
                 iou=0.45,
                 end2end=False):
        self.conf = conf
        self.iou = iou
        self.end2end = end2end
        self.model = ClassificationTrtPredictor(trt_file=trt_file, image_size=input_shape)

    def predict(self, image, class_names, use_preprocess=False):
        return self.model.predict(image, class_names=class_names, use_preprocess=use_preprocess)

    def feature(self, image, num_classes=1000, use_preprocess=False):
        return self.model.feature(image, num_classes=num_classes, use_preprocess=use_preprocess)

    @staticmethod
    def show(image, result):
        pass
